"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const common_tags_1 = require("common-tags");
const ast_helpers_1 = require("./ast_helpers");
const remove_decorators_1 = require("./remove_decorators");
describe('@ngtools/webpack transformers', () => {
    describe('decorator_remover', () => {
        it('should remove Angular decorators', () => {
            const input = common_tags_1.stripIndent `
        import { Component } from '@angular/core';

        @Component({
          selector: 'app-root',
          templateUrl: './app.component.html',
          styleUrls: ['./app.component.css']
        })
        export class AppComponent {
          title = 'app';
        }
      `;
            const output = common_tags_1.stripIndent `
        export class AppComponent {
          constructor() {
            this.title = 'app';
          }
        }
      `;
            const { program, compilerHost } = ast_helpers_1.createTypescriptContext(input);
            const transformer = remove_decorators_1.removeDecorators(() => true, () => program.getTypeChecker());
            const result = ast_helpers_1.transformTypescript(undefined, [transformer], program, compilerHost);
            expect(common_tags_1.oneLine `${result}`).toEqual(common_tags_1.oneLine `${output}`);
        });
        it('should not remove non-Angular decorators', () => {
            const input = common_tags_1.stripIndent `
        import { Component } from 'another-lib';

        @Component({
          selector: 'app-root',
          templateUrl: './app.component.html',
          styleUrls: ['./app.component.css']
        })
        export class AppComponent {
          title = 'app';
        }
      `;
            const output = `
        import * as tslib_1 from "tslib";
        import { Component } from 'another-lib';
        let AppComponent = class AppComponent {
            constructor() {
                this.title = 'app';
            }
        };
        AppComponent = tslib_1.__decorate([
            Component({
                selector: 'app-root',
                templateUrl: './app.component.html',
                styleUrls: ['./app.component.css']
            })
        ], AppComponent);
        export { AppComponent };
      `;
            const { program, compilerHost } = ast_helpers_1.createTypescriptContext(input);
            const transformer = remove_decorators_1.removeDecorators(() => true, () => program.getTypeChecker());
            const result = ast_helpers_1.transformTypescript(undefined, [transformer], program, compilerHost);
            expect(common_tags_1.oneLine `${result}`).toEqual(common_tags_1.oneLine `${output}`);
        });
        it('should remove imports for identifiers within the decorator', () => {
            const input = common_tags_1.stripIndent `
        import { Component } from '@angular/core';
        import { ChangeDetectionStrategy } from '@angular/core';

        @Component({
          selector: 'app-root',
          changeDetection: ChangeDetectionStrategy.OnPush,
          templateUrl: './app.component.html',
          styleUrls: ['./app.component.css']
        })
        export class AppComponent {
          title = 'app';
        }
      `;
            const output = common_tags_1.stripIndent `
        export class AppComponent {
          constructor() {
            this.title = 'app';
          }
        }
      `;
            const { program, compilerHost } = ast_helpers_1.createTypescriptContext(input);
            const transformer = remove_decorators_1.removeDecorators(() => true, () => program.getTypeChecker());
            const result = ast_helpers_1.transformTypescript(undefined, [transformer], program, compilerHost);
            expect(common_tags_1.oneLine `${result}`).toEqual(common_tags_1.oneLine `${output}`);
        });
        it('should not remove imports from types that are still used', () => {
            const input = common_tags_1.stripIndent `
        import { Component, EventEmitter } from '@angular/core';

        @Component({
          selector: 'app-root',
          changeDetection: ChangeDetectionStrategy.OnPush,
          templateUrl: './app.component.html',
          styleUrls: ['./app.component.css']
        })
        export class AppComponent {
          notify: EventEmitter<string> = new EventEmitter<string>();
          title = 'app';
        }
      `;
            const output = common_tags_1.stripIndent `
        import { EventEmitter } from '@angular/core';

        export class AppComponent {
          constructor() {
            this.notify = new EventEmitter();
            this.title = 'app';
          }
        }
      `;
            const { program, compilerHost } = ast_helpers_1.createTypescriptContext(input);
            const transformer = remove_decorators_1.removeDecorators(() => true, () => program.getTypeChecker());
            const result = ast_helpers_1.transformTypescript(undefined, [transformer], program, compilerHost);
            expect(common_tags_1.oneLine `${result}`).toEqual(common_tags_1.oneLine `${output}`);
        });
    });
});
//# sourceMappingURL=/users/dan/projects/angular-cli/src/transformers/remove_decorators.spec.js.map